@@ -3,9 +3,8 @@
     Copyright (C) 1997,1998  Matt Kimball
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,7 +22,8 @@
     Released under GPL, as above.
 */
 
-#include <config.h>
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/select.h>
@@ -32,8 +32,14 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+
+#ifndef __APPLE__
 #define BIND_8_COMPAT
+#endif
 #include <arpa/nameser.h>
+#ifdef HAVE_ARPA_NAMESER_COMPAT_H
+#include <arpa/nameser_compat.h>
+#endif
 #include <netdb.h>
 #include <resolv.h>
 #include <unistd.h>
@@ -49,6 +55,17 @@
 #include "dns.h"
 #include "net.h"
 
+#ifdef ENABLE_IPV6
+#ifdef __GLIBC__
+#define NSCOUNT6 myres._u._ext.nscount6
+#define NSSOCKADDR6(i) (myres._u._ext.nsaddrs[i])
+#else
+#define NSCOUNT6 myres.nscount
+#define NSSOCKADDR6(i) (&(myres._u._ext.ext->nsaddrs[i].sin6))
+#endif
+#endif
+
+
 #ifdef NO_STRERROR
 extern int sys_nerr;
 extern char *sys_errlist[];
@@ -165,7 +182,7 @@ char *rrtypes[] = {
 
 
 /* Please don't use a trailing comma in enumerations: It doesn't
-   work on all compilers -- REW */
+   work on all compilers */
 enum {
    RR_UNKNOWN,
    RR_QUERY,
@@ -214,8 +231,7 @@ typedef struct {
 #if 0
 
 /* The execution order inside an expression is undefined! That means that
-   this might work, but then again, it might not... Sun Lint pointed this 
-   one out...*/
+   this might work, but then again, it might not... */
 
 #define sucknetword(x) (((word)*(x) << 8) | (((x)+= 2)[-1]))
 #define sucknetshort(x) (((short)*(x) << 8) | (((x)+= 2)[-1]))
@@ -253,6 +269,9 @@ struct logline *lastlog = NULL;
 
 ip_t alignedip;
 ip_t localhost;
+#ifdef ENABLE_IPV6
+ip_t localhost6;
+#endif
 
 double sweeptime;
 
@@ -289,6 +308,9 @@ struct sockaddr_in * from4 = (struct sockaddr_in *) &from_sastruct;
 struct sockaddr * from = (struct sockaddr *) &from_sastruct;
 
 int resfd;
+#ifdef ENABLE_IPV6
+int resfd6;
+#endif
 socklen_t fromlen = sizeof from_sastruct;
 
 char tempstring[16384+1+1];
@@ -300,6 +322,17 @@ char nullstring[] = "";
 
 int use_dns = 1;
 
+#ifdef res_ninit
+#define MY_RES_INIT() res_ninit(&myres);
+#define RES_MKQUERY(a, b, c, d, e, f, g, h, i) \
+    res_nmkquery(&myres, a, b, c, d, e, f, g, h, i)
+struct __res_state myres;
+#else
+#define MY_RES_INIT() res_init();
+#define RES_MKQUERY(a, b, c, d, e, f, g, h, i) \
+    res_mkquery(a, b, c, d, e, f, g, h, i)
+#define myres _res
+#endif
 
 /* Code */
 #ifdef CorruptCheck
@@ -351,12 +384,12 @@ void statfree(void *p)
   if (p) {
     if (*((dword *)p - HEAD_SLACK) == 0) {
       fprintf(stderr,"ERROR: Attempt to free pointer twice.\n");
-      *(int*)0=0;
+      abort();
       exit(-1);
     } else {
       if (*((dword *)p - HEAD_SLACK) > 8192) {
 	fprintf (stderr,"ERROR: Corrupted free() buffer. (header)\n");
-	*(int*)0=0;
+        abort();
 	exit(-1);
       }
 #ifdef CorruptCheck
@@ -365,7 +398,7 @@ void statfree(void *p)
 	  (*(byte *)((char *)p + (*((dword *)p - 1)) + sizeof(byte) * 2) != 0xbe) ||
 	  (*(byte *)((char *)p + (*((dword *)p - 1)) + sizeof(byte) * 3) != 0xef)) {
 	fprintf(stderr,"ERROR: Corrupted free() buffer. (footer)\n");
-	*(int*)0=0;
+	abort();
 	exit(-1);
       }
 #endif
@@ -434,9 +467,9 @@ void clearset(fd_set *set)
 char *strlongip(ip_t * ip)
 {
 #ifdef ENABLE_IPV6
-  static char buf[INET6_ADDRSTRLEN];
+  static char addrstr[INET6_ADDRSTRLEN];
 
-  return (char *) inet_ntop( af, ip, buf, sizeof buf );
+  return (char *) inet_ntop( af, ip, addrstr, sizeof addrstr );
 #else
   return inet_ntoa( *ip );
 #endif
@@ -453,7 +486,7 @@ int longipstr( char *s, ip_t *dst, int af )
 }
 
 
-struct hostent * dns_forward(char *name)
+struct hostent * dns_forward(const char *name)
 {
   struct hostent *host;
 
@@ -468,33 +501,64 @@ int dns_waitfd(void)
 {
   return resfd;
 }
+#ifdef ENABLE_IPV6
+int dns_waitfd6(void)
+{
+  return resfd6;
+}
+#endif
 
 
 void dns_open(void)
 {
   int option,i;
 
-  res_init();
-  if (!_res.nscount) {
+  if (!dns) return;
+  MY_RES_INIT();
+  if (!myres.nscount) {
     fprintf(stderr,"No nameservers defined.\n");
     exit(-1);
   }
-  _res.options|= RES_RECURSE | RES_DEFNAMES | RES_DNSRCH;
-  for (i = 0;i < _res.nscount;i++)
-    _res.nsaddr_list[i].sin_family = AF_INET;
+  myres.options|= RES_RECURSE | RES_DEFNAMES | RES_DNSRCH;
   resfd = socket(AF_INET, SOCK_DGRAM, 0);
   if (resfd == -1) {
-    fprintf(stderr,"Unable to allocate socket for nameserver communication: %s\n",
+    fprintf(stderr,
+            "Unable to allocate IPv4 socket for nameserver communication: %s\n",
 	    strerror(errno));
     exit(-1);
   }
+#ifdef ENABLE_IPV6
+  resfd6 = socket(AF_INET6, SOCK_DGRAM, 0);
+  if (resfd6 == -1) {
+    // consider making removing this warning. For now leave it in to see 
+    // new code activated. -- REW
+    fprintf(stderr,
+            "Unable to allocate IPv6 socket for nameserver communication: %s\n",
+	    strerror(errno));
+    //    exit(-1);
+  }
+#endif
   option = 1;
   if (setsockopt(resfd,SOL_SOCKET,SO_BROADCAST,(char *)&option,sizeof(option))) {
-    fprintf(stderr,"Unable to setsockopt() on nameserver communication socket: %s\n",
+    fprintf(stderr,
+            "Unable to setsockopt() on IPv4 nameserver communication socket: %s\n",
 	    strerror(errno));
     exit(-1);
   }
+#ifdef ENABLE_IPV6
+  if (resfd6 > 0) {
+    if (setsockopt(resfd6,SOL_SOCKET,SO_BROADCAST,(char *)&option,sizeof(option))) {
+      fprintf(stderr,
+	      "Unable to setsockopt() on IPv6 nameserver communication socket: %s\n",
+	      strerror(errno));
+      exit(-1);
+    }
+  }
+#endif
   longipstr( "127.0.0.1", &localhost, AF_INET );
+#ifdef ENABLE_IPV6
+  longipstr( "::1", &localhost6, AF_INET6 );
+#endif
   aseed = time(NULL) ^ (time(NULL) << 3) ^ (dword)getpid();
   for (i = 0;i < BashSize;i++) {
     idbash[i] = NULL;
@@ -864,18 +928,30 @@ void dorequest(char *s,int type,word id)
 {
   packetheader *hp;
   int r,i;
-  int buf[(MaxPacketsize/sizeof (int))+1];
+  unsigned char buf[MaxPacketsize];
 
-  r = res_mkquery(QUERY,s,C_IN,type,NULL,0,NULL,(unsigned char*)buf,MaxPacketsize);
+  r = RES_MKQUERY(QUERY,s,C_IN,type,NULL,0,NULL,(unsigned char*)buf,MaxPacketsize);
   if (r == -1) {
     restell("Resolver error: Query too large.");
     return;
   }
   hp = (packetheader *)buf;
   hp->id = id;	/* htons() deliberately left out (redundant) */
-  for (i = 0;i < _res.nscount;i++)
-    (void)sendto(resfd,buf,r,0,(struct sockaddr *)&_res.nsaddr_list[i],
-		 sizeof(struct sockaddr));
+#ifdef ENABLE_IPV6
+  if (resfd6 > 0) {
+    for (i = 0;i < myres.nscount;i++) {
+      if (!NSSOCKADDR6(i))
+	continue;
+      if (NSSOCKADDR6(i)->sin6_family == AF_INET6)
+	(void)sendto(resfd6,buf,r,0,(struct sockaddr *) NSSOCKADDR6(i),
+		     sizeof(struct sockaddr_in6));
+    }
+  }
+#endif
+  for (i = 0;i < myres.nscount;i++)
+    if (myres.nsaddr_list[i].sin_family == AF_INET)
+      (void)sendto(resfd,buf,r,0,(struct sockaddr *)&myres.nsaddr_list[i],
+		   sizeof(struct sockaddr));
 }
 
 void resendrequest(struct resolve *rp,int type)
@@ -883,7 +959,7 @@ void resendrequest(struct resolve *rp,int type)
   if (type == T_A) {
     dorequest(rp->hostname,type,rp->id);
     if (debug) {
-      sprintf(tempstring,"Resolver: Sent reverse authentication request for \"%s\".",
+      snprintf(tempstring, sizeof(tempstring), "Resolver: Sent reverse authentication request for \"%s\".",
 	      rp->hostname);
       restell(tempstring);
     }
@@ -904,7 +980,7 @@ void resendrequest(struct resolve *rp,int type)
     }
     dorequest(tempstring,type,rp->id);
     if (debug) {
-      sprintf(tempstring,"Resolver: Sent domain lookup request for \"%s\".",
+      snprintf(tempstring, sizeof(tempstring), "Resolver: Sent domain lookup request for \"%s\".",
 	      strlongip( &(rp->ip) ));
       restell(tempstring);
     }
@@ -940,7 +1016,7 @@ void passrp(struct resolve *rp,long ttl)
   rp->expiretime = sweeptime + (double)ttl;
   untieresolve(rp);
   if (debug) {
-    sprintf(tempstring,"Resolver: Lookup successful: %s\n",rp->hostname);
+    snprintf(tempstring, sizeof(tempstring), "Resolver: Lookup successful: %s\n",rp->hostname);
     restell(tempstring);
   }
 }
@@ -997,7 +1073,7 @@ void parserespacket(byte *s, int l)
   case NOERROR:
     if (hp->ancount) {
       if (debug) {
-	sprintf(tempstring,"Resolver: Received nameserver reply. (qd:%u an:%u ns:%u ar:%u)",
+	snprintf(tempstring, sizeof(tempstring), "Resolver: Received nameserver reply. (qd:%u an:%u ns:%u ar:%u)",
                 hp->qdcount,hp->ancount,hp->nscount,hp->arcount);
 	restell(tempstring);
       }
@@ -1037,14 +1113,14 @@ void parserespacket(byte *s, int l)
       namestring[strlen(stackstring)] = '\0';
       if (strcasecmp(stackstring,namestring)) {
 	if (debug) {
-	  sprintf(tempstring,"Resolver: Unknown query packet dropped. (\"%s\" does not match \"%s\")",
+	  snprintf(tempstring, sizeof(tempstring), "Resolver: Unknown query packet dropped. (\"%s\" does not match \"%s\")",
 		  stackstring,namestring);
 	  restell(tempstring);
 	}
 	return;
       }
       if (debug) {
-	sprintf(tempstring,"Resolver: Queried domain name: \"%s\"",namestring);
+	snprintf(tempstring, sizeof(tempstring), "Resolver: Queried domain name: \"%s\"",namestring);
 	restell(tempstring);
       }
       c+= r;
@@ -1055,7 +1131,7 @@ void parserespacket(byte *s, int l)
       qdatatype = sucknetword(c);
       qclass = sucknetword(c);
       if (qclass != C_IN) {
-	sprintf(tempstring,"Resolver error: Received unsupported query class: %u (%s)",
+	snprintf(tempstring, sizeof(tempstring), "Resolver error: Received unsupported query class: %u (%s)",
                 qclass,qclass < ClasstypeCount ? classtypes[qclass] :
 		classtypes[ClasstypeCount]);
 	restell(tempstring);
@@ -1069,7 +1145,7 @@ void parserespacket(byte *s, int l)
 	  }
 	break;
       default:
-	sprintf(tempstring,"Resolver error: Received unimplemented query type: %u (%s)",
+	snprintf(tempstring, sizeof(tempstring), "Resolver error: Received unimplemented query type: %u (%s)",
 		qdatatype,qdatatype < ResourcetypeCount ?
 		resourcetypes[qdatatype] : resourcetypes[ResourcetypeCount]);
 	restell(tempstring);
@@ -1091,7 +1167,7 @@ void parserespacket(byte *s, int l)
 	else
 	  usefulanswer = 1;
 	if (debug) {
-	  sprintf(tempstring,"Resolver: answered domain query: \"%s\"",namestring);
+	  snprintf(tempstring, sizeof(tempstring), "Resolver: answered domain query: \"%s\"",namestring);
 	  restell(tempstring);
 	}
 	c+= r;
@@ -1104,10 +1180,10 @@ void parserespacket(byte *s, int l)
 	ttl = sucknetlong(c);
 	rdatalength = sucknetword(c);
 	if (class != qclass) {
-	  sprintf(tempstring,"query class: %u (%s)",qclass,qclass < ClasstypeCount ?
+	  snprintf(tempstring, sizeof(tempstring), "query class: %u (%s)",qclass,qclass < ClasstypeCount ?
 		  classtypes[qclass] : classtypes[ClasstypeCount]);
 	  restell(tempstring);
-	  sprintf(tempstring,"rr class: %u (%s)",class,class < ClasstypeCount ?
+	  snprintf(tempstring, sizeof(tempstring), "rr class: %u (%s)",class,class < ClasstypeCount ?
 		  classtypes[class] : classtypes[ClasstypeCount]);
 	  restell(tempstring);
 	  restell("Resolver error: Answered class does not match queried class.");
@@ -1123,20 +1199,20 @@ void parserespacket(byte *s, int l)
 	}
 	if (datatype == qdatatype || datatype == T_CNAME) {
 	  if (debug) {
-	    sprintf(tempstring,"Resolver: TTL: %s",strtdiff(sendstring,ttl));
+	    snprintf(tempstring, sizeof(tempstring), "Resolver: TTL: %s",strtdiff(sendstring,ttl));
 	    restell(tempstring);
 	  }
 	  if (usefulanswer)
 	    switch (datatype) {
 	    case T_A:
 	      if (rdatalength != 4) {
-		sprintf(tempstring,"Resolver error: Unsupported rdata format for \"A\" type. (%u bytes)",
+		snprintf(tempstring, sizeof(tempstring), "Resolver error: Unsupported rdata format for \"A\" type. (%u bytes)",
 			rdatalength);
 		restell(tempstring);
 		return;
 	      }
 	      if ( addrcmp( (void *) &(rp->ip), (void *) c, af ) == 0 ) {
-		sprintf(tempstring,"Resolver: Reverse authentication failed: %s != ",
+		snprintf(tempstring, sizeof(tempstring), "Resolver: Reverse authentication failed: %s != ",
 			strlongip( &(rp->ip) ));
 		addrcpy( (void *) &alignedip, (void *) c, af );
 		strcat(tempstring,strlongip( &alignedip ));
@@ -1144,7 +1220,7 @@ void parserespacket(byte *s, int l)
 		res_hostipmismatch++;
 		failrp(rp);
 	      } else {
-		sprintf(tempstring,"Resolver: Reverse authentication complete: %s == \"%s\".",
+		snprintf(tempstring, sizeof(tempstring), "Resolver: Reverse authentication complete: %s == \"%s\".",
 			strlongip( &(rp->ip) ),nonull(rp->hostname));
 		restell(tempstring);
 		res_reversesuccess++;
@@ -1161,7 +1237,7 @@ void parserespacket(byte *s, int l)
 		return;
 	      }
 	      if (debug) {
-		sprintf(tempstring,"Resolver: Answered domain: \"%s\"",namestring);
+		snprintf(tempstring, sizeof(tempstring), "Resolver: Answered domain: \"%s\"",namestring);
 		restell(tempstring);
 	      }
 	      if (r > HostnameLength) {
@@ -1186,14 +1262,14 @@ void parserespacket(byte *s, int l)
 	      }
 	      break;
 	    default:
-	      sprintf(tempstring,"Resolver error: Received unimplemented data type: %u (%s)",
+	      snprintf(tempstring, sizeof(tempstring), "Resolver error: Received unimplemented data type: %u (%s)",
 		      datatype,datatype < ResourcetypeCount ?
 		      resourcetypes[datatype] : resourcetypes[ResourcetypeCount]);
 	      restell(tempstring);
 	    }
 	} else {
 	  if (debug) {
-	    sprintf(tempstring,"Resolver: Ignoring resource type %u. (%s)",
+	    snprintf(tempstring, sizeof(tempstring), "Resolver: Ignoring resource type %u. (%s)",
 		    datatype,datatype < ResourcetypeCount ?
 		    resourcetypes[datatype] : resourcetypes[ResourcetypeCount]);
 	    restell(tempstring);
@@ -1211,7 +1287,7 @@ void parserespacket(byte *s, int l)
     failrp(rp);
     break;
   default:
-    sprintf(tempstring,"Resolver: Received error response %u. (%s)",
+    snprintf(tempstring, sizeof(tempstring), "Resolver: Received error response %u. (%s)",
 	    getheader_rcode(hp),getheader_rcode(hp) < ResponsecodeCount ?
 	    responsecodes[getheader_rcode(hp)] : responsecodes[ResponsecodeCount]);
     restell(tempstring);
@@ -1230,28 +1306,75 @@ void dns_ack(void)
     /* Check to see if this server is actually one we sent to */
     if ( addrcmp( (void *) &(from4->sin_addr), (void *) &localhost,
                   (int) AF_INET ) == 0 ) {
-      for (i = 0;i < _res.nscount;i++)
-	if ( addrcmp( (void *) &(_res.nsaddr_list[i].sin_addr),
+      for (i = 0;i < myres.nscount;i++)
+	if ( addrcmp( (void *) &(myres.nsaddr_list[i].sin_addr),
 		      (void *) &(from4->sin_addr), (int) AF_INET ) == 0 ||
-	     addrcmp( (void *) &(_res.nsaddr_list[i].sin_addr),
+	     addrcmp( (void *) &(myres.nsaddr_list[i].sin_addr),
 		      (void *) &unspec_addr, (int) AF_INET ) == 0 )	/* 0.0.0.0 replies as 127.0.0.1 */
 	  break;
     } else
-      for (i = 0;i < _res.nscount;i++)
-	if ( addrcmp( (void *) &(_res.nsaddr_list[i].sin_addr),
+      for (i = 0;i < myres.nscount;i++)
+	if ( addrcmp( (void *) &(myres.nsaddr_list[i].sin_addr),
 		      (void *) &(from4->sin_addr), AF_INET ) == 0 )
 	  break;
-    if (i == _res.nscount) {
-      sprintf(tempstring,"Resolver error: Received reply from unknown source: %s",
+    if (i == myres.nscount) {
+      snprintf(tempstring, sizeof(tempstring), "Resolver error: Received reply from unknown source: %s",
 	      inet_ntoa(from4->sin_addr ));
       restell(tempstring);
     } else
       parserespacket((byte *)resrecvbuf,r);
   } else {
-    sprintf(tempstring,"Resolver: Socket error: %s",strerror(errno));
+    snprintf(tempstring, sizeof(tempstring), "Resolver: Socket error: %s",strerror(errno));
     restell(tempstring);
   }
 }
+#ifdef ENABLE_IPV6
+void dns_ack6(void)
+{
+  int r,i;
+  static char addrstr[INET6_ADDRSTRLEN];
+
+  // Probably not necessary. -- REW
+  if (resfd6 < 0) return; 
+
+  r = recvfrom(resfd6,(byte *)resrecvbuf,MaxPacketsize,0,
+               from, &fromlen);
+  if (r > 0) {
+    /* Check to see if this server is actually one we sent to */
+    if ( addrcmp( (void *) &(from6->sin6_addr), (void *) &localhost6,
+                  (int) AF_INET6 ) == 0 ) {
+      for (i = 0;i < myres.nscount;i++) {
+        if (!NSSOCKADDR6(i))
+          continue;
+
+	if ( addrcmp( (void *) &(NSSOCKADDR6(i)->sin6_addr),
+		      (void *) &(from6->sin6_addr), (int) AF_INET6 ) == 0 ||
+	     addrcmp( (void *) &(NSSOCKADDR6(i)->sin6_addr),
+		      (void *) &unspec_addr, (int) AF_INET6 ) == 0 )	/* 0.0.0.0 replies as 127.0.0.1 */
+	  break;
+      }
+    } else
+      for (i = 0;i < myres.nscount;i++) {
+        if (!NSSOCKADDR6(i))
+          continue;
+	if ( addrcmp( (void *) &(NSSOCKADDR6(i)->sin6_addr),
+		      (void *) &(from6->sin6_addr), AF_INET6 ) == 0 )
+	  break;
+      }
+    if (i == myres.nscount) {
+      snprintf(tempstring, sizeof(tempstring), 
+	       "Resolver error: Received reply from unknown source: %s",
+	       inet_ntop( AF_INET6, &(from6->sin6_addr), addrstr,
+			  sizeof addrstr ));
+      restell(tempstring);
+    } else
+      parserespacket((byte *)resrecvbuf,r);
+  } else {
+    snprintf(tempstring, sizeof(tempstring), "Resolver: Socket error: %s",strerror(errno));
+    restell(tempstring);
+  }
+}
+#endif
 
 
 int istime(double x,double *sinterval)
@@ -1277,7 +1400,7 @@ void dns_events(double *sinterval)
     case STATE_FINISHED:	/* TTL has expired */
     case STATE_FAILED:	/* Fake TTL has expired */
       if (debug) {
-	sprintf(tempstring,"Resolver: Cache record for \"%s\" (%s) has expired. (state: %u)  Marked for expire at: %g, time: %g.",
+	snprintf(tempstring, sizeof(tempstring), "Resolver: Cache record for \"%s\" (%s) has expired. (state: %u)  Marked for expire at: %g, time: %g.",
                 nonull(rp->hostname), strlongip( &(rp->ip) ), 
 		rp->state, rp->expiretime, sweeptime);
 	restell(tempstring);
@@ -1321,14 +1444,14 @@ char *dns_lookup2(ip_t * ip)
     if ((rp->state == STATE_FINISHED) || (rp->state == STATE_FAILED)) {
       if ((rp->state == STATE_FINISHED) && (rp->hostname)) {
 	if (debug) {
-	  sprintf(tempstring,"Resolver: Used cached record: %s == \"%s\".\n",
+	  snprintf(tempstring, sizeof(tempstring), "Resolver: Used cached record: %s == \"%s\".\n",
 		  strlongip(ip),rp->hostname);
 	  restell(tempstring);
 	}
 	return rp->hostname;
       } else {
 	if (debug) {
-	  sprintf(tempstring,"Resolver: Used failed record: %s == ???\n",
+	  snprintf(tempstring, sizeof(tempstring), "Resolver: Used failed record: %s == ???\n",
 		  strlongip(ip));
 	  restell(tempstring);
 	}
@@ -1355,9 +1478,9 @@ char *dns_lookup(ip_t * ip)
 {
   char *t;
 
-  if (!dns) return strlongip (ip);
-  t = dns_lookup2 (ip);
-  return (t&&use_dns)?t:strlongip(ip);
+  if (!dns) return NULL;
+  t = dns_lookup2(ip);
+  return (t && use_dns) ? t : NULL;
 }
 
 #ifdef ENABLE_IPV6
