@@ -3,9 +3,8 @@
     Copyright (C) 1997,1998  Matt Kimball
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,12 +14,9 @@
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-    
-   1999-08-13 ok Olav@okvittem.priv.no  added -psize
-
 */
 
-#include <config.h>
+#include "config.h"
 
 #if defined(HAVE_SYS_XTI_H)
 #include <sys/xti.h>
@@ -29,9 +25,14 @@
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
+#include <sys/select.h>
 #include <netinet/in.h>
 #include <memory.h>
 #include <unistd.h>
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
@@ -54,6 +55,29 @@ struct ICMPHeader {
   uint16 sequence;
 };
 
+/* Structure of an UDP header.  */
+struct UDPHeader {
+  uint16 srcport;
+  uint16 dstport;
+  uint16 length;
+  uint16 checksum;
+};
+
+/* Structure of an TCP header, as far as we need it.  */
+struct TCPHeader {
+  uint16 srcport;
+  uint16 dstport;
+  uint32 seq;
+};
+
+/* Structure of an IPv4 UDP pseudoheader.  */
+struct UDPv4PHeader {
+  uint32 saddr;
+  uint32 daddr;
+  uint8 zero;
+  uint8 protocol;
+  uint16 len;
+};
 
 /*  Structure of an IP header.  */
 struct IPHeader {
@@ -77,6 +101,7 @@ struct IPHeader {
 #define ICMP_TSTAMPREPLY	14
 
 #define ICMP_TIME_EXCEEDED	11
+#define ICMP_UNREACHABLE        3
 
 #ifndef SOL_IP
 #define SOL_IP 0
@@ -96,13 +121,15 @@ struct nethost {
   int avg;	/* average:  addByMin */
   int gmean;	/* geometirc mean: addByMin */
   int jitter;	/* current jitter, defined as t1-t0 addByMin */
-//int jbest;	/* min jitter, of cause it is 0, not needed */
+/*int jbest;*/	/* min jitter, of cause it is 0, not needed */
   int javg;	/* avg jitter */
   int jworst;	/* max jitter */
   int jinta;	/* estimated variance,? rfc1889's "Interarrival Jitter" */
   int transit;
   int saved[SAVED_PINGS];
   int saved_seq_offset;
+  struct mplslen mpls;
+  struct mplslen mplss[MAXPATH];
 };
 
 
@@ -111,12 +138,13 @@ struct sequence {
   int transit;
   int saved_seq;
   struct timeval time;
+  int socket;
 };
 
 
 /* Configuration parameter: How many queries to unknown hosts do we
    send? (This limits the amount of traffic generated if a host is not
-   reachable) -- REW */
+   reachable) */
 #define MAX_UNKNOWN_HOSTS 5
 
 
@@ -131,8 +159,12 @@ static struct timeval reset = { 0, 0 };
 
 int    timestamp;
 int    sendsock4;
+int    sendsock4_icmp;
+int    sendsock4_udp;
 int    recvsock4;
 int    sendsock6;
+int    sendsock6_icmp;
+int    sendsock6_udp;
 int    recvsock6;
 int    sendsock;
 int    recvsock;
@@ -155,7 +187,7 @@ struct sockaddr_in * rsa4 = (struct sockaddr_in *) &remotesockaddr_struct;
 ip_t * sourceaddress;
 ip_t * remoteaddress;
 
-/* XXX How do I code this to be IPV6 compatible??? -- REW */
+/* XXX How do I code this to be IPV6 compatible??? */
 #ifdef ENABLE_IPV6
 char localaddr[INET6_ADDRSTRLEN];
 #else
@@ -175,7 +207,9 @@ static int packetsize;		/* packet size used by ping */
 extern int bitpattern;		/* packet bit pattern used by ping */
 extern int tos;			/* type of service set in ping packet*/
 extern int af;			/* address family of remote target */
-
+extern int mtrtype;		/* type of query packet used */
+extern int remoteport;          /* target port for TCP tracing */
+extern int timeout;             /* timeout for TCP connections */
 
 /* return the number of microseconds to wait before sending the next
    ping */
@@ -206,15 +240,34 @@ int checksum(void *data, int sz)
 }
 
 
-int new_sequence(int index) 
+/* Prepend pseudoheader to the udp datagram and calculate checksum */
+int udp_checksum(void *pheader, void *udata, int psize, int dsize)
 {
-  static int next_sequence = 0;
-  int seq;
+  unsigned int tsize = psize + dsize;
+  char csumpacket[tsize];
+  memset(csumpacket, (unsigned char) abs(bitpattern), abs(tsize));
 
-  seq = next_sequence++;
-  if (next_sequence >= MaxSequence)
-    next_sequence = 0;
+  struct UDPv4PHeader *prepend = (struct UDPv4PHeader *) csumpacket;
+  struct UDPv4PHeader *udppheader = (struct UDPv4PHeader *) pheader;
+  prepend->saddr = udppheader->saddr;
+  prepend->daddr = udppheader->daddr;
+  prepend->zero = 0;
+  prepend->protocol = udppheader->protocol;
+  prepend->len = udppheader->len;
 
+  struct UDPHeader *content = (struct UDPHeader *)(csumpacket + psize);
+  struct UDPHeader *udpdata = (struct UDPHeader *) udata;
+  content->srcport = udpdata->srcport;
+  content->dstport = udpdata->dstport;
+  content->length = udpdata->length;
+  content->checksum = udpdata->checksum;
+
+  return checksum(csumpacket,tsize);
+}
+
+
+void save_sequence(int index, int seq)
+{
   sequence[seq].index = index;
   sequence[seq].transit = 1;
   sequence[seq].saved_seq = ++host[index].xmit;
@@ -225,18 +278,146 @@ int new_sequence(int index)
     host[index].up = 0;
   host[index].sent = 1;
   net_save_xmit(index);
-  
+}
+
+int new_sequence(int index)
+{
+  static int next_sequence = MinSequence;
+  int seq;
+
+  seq = next_sequence++;
+  if (next_sequence >= MaxSequence)
+    next_sequence = MinSequence;
+
+  save_sequence(index, seq);
+
   return seq;
 }
 
+/*  Attempt to connect to a TCP port with a TTL */
+void net_send_tcp(int index)
+{
+  int ttl, s;
+  int opt = 1;
+  int port;
+  struct sockaddr_storage local;
+  struct sockaddr_storage remote;
+  struct sockaddr_in *local4 = (struct sockaddr_in *) &local;
+  struct sockaddr_in6 *local6 = (struct sockaddr_in6 *) &local;
+  struct sockaddr_in *remote4 = (struct sockaddr_in *) &remote;
+  struct sockaddr_in6 *remote6 = (struct sockaddr_in6 *) &remote;
+  socklen_t len;
+
+  ttl = index + 1;
+
+  s = socket(af, SOCK_STREAM, 0);
+  if (s < 0) {
+    display_clear();
+    perror("socket()");
+    exit(EXIT_FAILURE);
+  }
+
+  memset(&local, 0, sizeof (local));
+  memset(&remote, 0, sizeof (remote));
+  local.ss_family = af;
+  remote.ss_family = af;
+
+  switch (af) {
+  case AF_INET:
+    addrcpy((void *) &local4->sin_addr, (void *) &ssa4->sin_addr, af);
+    addrcpy((void *) &remote4->sin_addr, (void *) remoteaddress, af);
+    remote4->sin_port = htons(remoteport);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    addrcpy((void *) &local6->sin6_addr, (void *) &ssa6->sin6_addr, af);
+    addrcpy((void *) &remote6->sin6_addr, (void *) remoteaddress, af);
+    remote6->sin6_port = htons(remoteport);
+    break;
+#endif
+  }
+
+  if (bind(s, (struct sockaddr *) &local, sizeof (local))) {
+    display_clear();
+    perror("bind()");
+    exit(EXIT_FAILURE);
+  }
+
+  len = sizeof (local);
+  if (getsockname(s, (struct sockaddr *) &local, &len)) {
+    display_clear();
+    perror("getsockname()");
+    exit(EXIT_FAILURE);
+  }
+
+  opt = 1;
+  if (ioctl(s, FIONBIO, &opt)) {
+    display_clear();
+    perror("ioctl FIONBIO");
+    exit(EXIT_FAILURE);
+  }
+
+  switch (af) {
+  case AF_INET:
+    if (setsockopt(s, IPPROTO_IP, IP_TTL, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    if (setsockopt(s, IPPROTO_IP, IP_TOS, &tos, sizeof (tos))) {
+      display_clear();
+      perror("setsockopt IP_TOS");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    if (setsockopt(s, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#endif
+  }
+
+  switch (local.ss_family) {
+  case AF_INET:
+    port = ntohs(local4->sin_port);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    port = ntohs(local6->sin6_port);
+    break;
+#endif
+  default:
+    display_clear();
+    perror("unknown AF?");
+    exit(EXIT_FAILURE);
+  }
+
+  save_sequence(index, port);
+  gettimeofday(&sequence[port].time, NULL);
+  sequence[port].socket = s;
+
+  connect(s, (struct sockaddr *) &remote, sizeof (remote));
+}
 
 /*  Attempt to find the host at a particular number of hops away  */
 void net_send_query(int index) 
 {
+  if (mtrtype == IPPROTO_TCP) {
+    net_send_tcp(index);
+    return;
+  }
+
   /*ok  char packet[sizeof(struct IPHeader) + sizeof(struct ICMPHeader)];*/
   char packet[MAXPACKET];
   struct IPHeader *ip = (struct IPHeader *) packet;
-  struct ICMPHeader *icmp;
+  struct ICMPHeader *icmp = NULL;
+  struct UDPHeader *udp = NULL;
+  struct UDPv4PHeader *udpp = NULL;
+  uint16 mypid;
 
   /*ok  int packetsize = sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + datasize;*/
   int rv;
@@ -245,6 +426,11 @@ void net_send_query(int index)
 
   ttl = index + 1;
 
+#ifdef ENABLE_IPV6
+  /* offset for ipv6 checksum calculation */
+  int offset = 6;
+#endif
+
   if ( packetsize < MINPACKET ) packetsize = MINPACKET;
   if ( packetsize > MAXPACKET ) packetsize = MAXPACKET;
 
@@ -271,7 +457,7 @@ void net_send_query(int index)
   ip->id = 0;
   ip->frag = 0;    /* 1, if want to find mtu size? Min */
     ip->ttl = ttl;
-  ip->protocol = IPPROTO_ICMP;
+  ip->protocol = mtrtype;
   ip->check = 0;
 
   /* BSD needs the source address here, Linux & others do not... */
@@ -295,21 +481,69 @@ void net_send_query(int index)
 #endif
   }
 
-  icmp = (struct ICMPHeader *)(packet + iphsize);
-  icmp->type     = echotype;
-  icmp->code     = 0;
-  icmp->checksum = 0;
-  icmp->id       = getpid();
-  icmp->sequence = new_sequence(index);
-  icmp->checksum = checksum(icmp, abs(packetsize) - iphsize);
+  switch ( mtrtype ) {
+  case IPPROTO_ICMP:
+    icmp = (struct ICMPHeader *)(packet + iphsize);
+    icmp->type     = echotype;
+    icmp->code     = 0;
+    icmp->checksum = 0;
+    icmp->id       = getpid();
+    icmp->sequence = new_sequence(index);
+    icmp->checksum = checksum(icmp, abs(packetsize) - iphsize);
+    
+    gettimeofday(&sequence[icmp->sequence].time, NULL);
+    break;
 
-  switch ( af ) {
-  case AF_INET:
-    ip->check = checksum(packet, abs(packetsize));
+  case IPPROTO_UDP:
+    udp = (struct UDPHeader *)(packet + iphsize);
+    udp->checksum  = 0;
+    mypid = (uint16)getpid();
+    if (mypid < MinPort)
+      mypid += MinPort;
+
+    udp->srcport = htons(mypid);
+    udp->length = abs(packetsize) - iphsize;
+    if(!BSDfix)
+      udp->length = htons(udp->length);
+ 
+    udp->dstport = new_sequence(index);
+    gettimeofday(&sequence[udp->dstport].time, NULL);
+    udp->dstport = htons(udp->dstport);
     break;
   }
 
-  gettimeofday(&sequence[icmp->sequence].time, NULL);
+  switch ( af ) {
+  case AF_INET:
+    switch ( mtrtype ) {
+    case IPPROTO_UDP:
+      /* checksum is not mandatory. only calculate if we know ip->saddr */
+      if (ip->saddr) {
+        udpp = (struct UDPv4PHeader *)(malloc(sizeof(struct UDPv4PHeader)));
+        udpp->saddr = ip->saddr;
+        udpp->daddr = ip->daddr;
+        udpp->protocol = ip->protocol;
+        udpp->len = udp->length;
+        udp->checksum = udp_checksum(udpp, udp, sizeof(struct UDPv4PHeader), abs(packetsize) - iphsize);
+      }
+      break;
+    }
+
+    ip->check = checksum(packet, abs(packetsize));
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    switch ( mtrtype ) {
+    case IPPROTO_UDP:
+      /* kernel checksum calculation */
+      if ( setsockopt(sendsock, IPPROTO_IPV6, IPV6_CHECKSUM, &offset, sizeof(offset)) ) {
+        perror( "setsockopt IPV6_CHECKSUM" );
+        exit( EXIT_FAILURE);
+      }
+      break;
+    }
+    break;
+#endif
+  }
 
   rv = sendto(sendsock, packet, abs(packetsize), 0, 
 	      remotesockaddr, salen);
@@ -328,7 +562,7 @@ void net_send_query(int index)
 
 /*   We got a return on something we sent out.  Record the address and
      time.  */
-void net_process_ping(int seq, void * addr, struct timeval now) 
+void net_process_ping(int seq, struct mplslen mpls, void * addr, struct timeval now) 
 {
   int index;
   int totusec;
@@ -351,6 +585,11 @@ void net_process_ping(int seq, void * addr, struct timeval now)
     return;
   sequence[seq].transit = 0;
 
+  if (sequence[seq].socket > 0) {
+    close(sequence[seq].socket);
+    sequence[seq].socket = 0;
+  }
+
   index = sequence[seq].index;
 
   totusec = (now.tv_sec  - sequence[seq].time.tv_sec ) * 1000000 +
@@ -359,12 +598,14 @@ void net_process_ping(int seq, void * addr, struct timeval now)
 
   if ( addrcmp( (void *) &(host[index].addr),
 		(void *) &unspec_addr, af ) == 0 ) {
-    // should be out of if as addr can change
+    /* should be out of if as addr can change */
     addrcpy( (void *) &(host[index].addr), addrcopy, af );
+    host[index].mpls = mpls;
     display_rawhost(index, (void *) &(host[index].addr));
 
-  /* multi paths by Min */
+  /* multi paths */
     addrcpy( (void *) &(host[index].addrs[0]), addrcopy, af );
+    host[index].mplss[0] = mpls;
   } else {
     for( i=0; i<MAXPATH; ) {
       if( addrcmp( (void *) &(host[index].addrs[i]), (void *) &addrcopy,
@@ -376,8 +617,9 @@ void net_process_ping(int seq, void * addr, struct timeval now)
     if( addrcmp( (void *) &(host[index].addrs[i]), addrcopy, af ) != 0 && 
         i<MAXPATH ) {
       addrcpy( (void *) &(host[index].addrs[i]), addrcopy, af );
+      host[index].mplss[i] = mpls;
+      display_rawhost(index, (void *) &(host[index].addrs[i]));
     }
-  /* end multi paths */
   }
 
   host[index].jitter = totusec - host[index].last;
@@ -395,7 +637,6 @@ void net_process_ping(int seq, void * addr, struct timeval now)
    * at least in linux 2.4.x.
    *  safe guard 1) best[index]>=best[index-1] if index>0
    *             2) best >= average-20,000 usec (good number?)
-   *  Min
   if (index > 0) {
     if (totusec < host[index].best &&
        totusec>= host[index-1].best) host[index].best  = totusec;
@@ -410,10 +651,9 @@ void net_process_ping(int seq, void * addr, struct timeval now)
 	host[index].jworst = host[index].jitter;
 
   host[index].returned++;
-  /* begin addByMin do more stats */
   oldavg = host[index].avg;
   host[index].avg += (totusec - oldavg +.0) / host[index].returned;
-  host[index].var += (totusec - oldavg +.0) * (totusec - host[index].avg);
+  host[index].var += (totusec - oldavg +.0) * (totusec - host[index].avg) / 1000000;
 
   oldjavg = host[index].javg;
   host[index].javg += (host[index].jitter - oldjavg) / host[index].returned;
@@ -423,7 +663,6 @@ void net_process_ping(int seq, void * addr, struct timeval now)
   if ( host[index].returned > 1 )
   host[index].gmean = pow( (double) host[index].gmean, (host[index].returned-1.0)/host[index].returned )
 			* pow( (double) totusec, 1.0/host[index].returned );
-  /* end addByMin*/
   host[index].sent = 0;
   host[index].up = 1;
   host[index].transit = 0;
@@ -450,9 +689,16 @@ void net_process_return(void)
   socklen_t fromsockaddrsize;
   int num;
   struct ICMPHeader *header = NULL;
+  struct UDPHeader *udpheader = NULL;
+  struct TCPHeader *tcpheader = NULL;
   struct timeval now;
   ip_t * fromaddress = NULL;
-  int echoreplytype = 0, timeexceededtype = 0;
+  int echoreplytype = 0, timeexceededtype = 0, unreachabletype = 0;
+  int sequence = 0;
+
+  /* MPLS decoding */
+  struct mplslen mpls;
+  mpls.labels = 0;
 
   gettimeofday(&now, NULL);
   switch ( af ) {
@@ -461,6 +707,7 @@ void net_process_return(void)
     fromaddress = (ip_t *) &(fsa4->sin_addr);
     echoreplytype = ICMP_ECHOREPLY;
     timeexceededtype = ICMP_TIME_EXCEEDED;
+    unreachabletype = ICMP_UNREACHABLE;
     break;
 #ifdef ENABLE_IPV6
   case AF_INET6:
@@ -468,6 +715,7 @@ void net_process_return(void)
     fromaddress = (ip_t *) &(fsa6->sin6_addr);
     echoreplytype = ICMP6_ECHO_REPLY;
     timeexceededtype = ICMP6_TIME_EXCEEDED;
+    unreachabletype = ICMP6_DST_UNREACH;
     break;
 #endif
   }
@@ -490,41 +738,131 @@ void net_process_return(void)
     break;
 #endif
   }
-  if (header->type == echoreplytype) {
-    if(header->id != (uint16)getpid())
-      return;
 
-    net_process_ping (header->sequence, (void *) fromaddress, now);
-  } else if (header->type == timeexceededtype) {
-    switch ( af ) {
-    case AF_INET:
+  switch ( mtrtype ) {
+  case IPPROTO_ICMP:
+    if (header->type == echoreplytype) {
+      if(header->id != (uint16)getpid())
+        return;
+
+      sequence = header->sequence;
+    } else if (header->type == timeexceededtype) {
+      switch ( af ) {
+      case AF_INET:
+
+        if ((size_t) num < sizeof(struct IPHeader) + 
+                           sizeof(struct ICMPHeader) + 
+                           sizeof (struct IPHeader) + 
+                           sizeof (struct ICMPHeader))
+          return;
+        header = (struct ICMPHeader *)(packet + sizeof (struct IPHeader) + 
+                                                sizeof (struct ICMPHeader) + 
+                                                sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
 
-      if ((size_t) num < sizeof(struct IPHeader) + 
-                         sizeof(struct ICMPHeader) + 
-                         sizeof (struct IPHeader) + 
-                         sizeof (struct ICMPHeader))
-        return;
-      header = (struct ICMPHeader *)(packet + sizeof (struct IPHeader) + 
-                                              sizeof (struct ICMPHeader) + 
-                                              sizeof (struct IPHeader));
-    break;
-#ifdef ENABLE_IPV6
-    case AF_INET6:
-      if ( num < sizeof (struct ICMPHeader) + 
-                 sizeof (struct ip6_hdr) + sizeof (struct ICMPHeader) )
-        return;
-      header = (struct ICMPHeader *) ( packet + 
-                                       sizeof (struct ICMPHeader) +
-                                       sizeof (struct ip6_hdr) );
       break;
+#ifdef ENABLE_IPV6
+      case AF_INET6:
+        if ( num < sizeof (struct ICMPHeader) + 
+                   sizeof (struct ip6_hdr) + sizeof (struct ICMPHeader) )
+          return;
+        header = (struct ICMPHeader *) ( packet + 
+                                         sizeof (struct ICMPHeader) +
+                                         sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
+        break;
 #endif
+      }
+  
+      if (header->id != (uint16)getpid())
+        return;
+  
+      sequence = header->sequence;
     }
+    break;
+  
+  case IPPROTO_UDP:
+    if (header->type == timeexceededtype || header->type == unreachabletype) {
+      switch ( af ) {
+      case AF_INET:
 
-    if (header->id != (uint16)getpid())
-      return;
+        if ((size_t) num < sizeof(struct IPHeader) +
+                           sizeof(struct ICMPHeader) +
+                           sizeof (struct IPHeader) +
+                           sizeof (struct UDPHeader))
+          return;
+        udpheader = (struct UDPHeader *)(packet + sizeof (struct IPHeader) +
+                                                  sizeof (struct ICMPHeader) +
+                                                  sizeof (struct IPHeader));
 
-    net_process_ping(header->sequence, (void *)fromaddress, now);
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
+      break;
+#ifdef ENABLE_IPV6
+      case AF_INET6:
+        if ( num < sizeof (struct ICMPHeader) +
+                   sizeof (struct ip6_hdr) + sizeof (struct UDPHeader) )
+          return;
+        udpheader = (struct UDPHeader *) ( packet +
+                                           sizeof (struct ICMPHeader) +
+                                           sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
+        break;
+#endif
+      }
+      sequence = ntohs(udpheader->dstport);
+    }
+    break;
+
+  case IPPROTO_TCP:
+    if (header->type == timeexceededtype || header->type == unreachabletype) {
+      switch ( af ) {
+      case AF_INET:
+
+        if ((size_t) num < sizeof(struct IPHeader) +
+                           sizeof(struct ICMPHeader) +
+                           sizeof (struct IPHeader) +
+                           sizeof (struct TCPHeader))
+          return;
+        tcpheader = (struct TCPHeader *)(packet + sizeof (struct IPHeader) +
+                                                  sizeof (struct ICMPHeader) +
+                                                  sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
+      break;
+#ifdef ENABLE_IPV6
+      case AF_INET6:
+        if ( num < sizeof (struct ICMPHeader) +
+                   sizeof (struct ip6_hdr) + sizeof (struct TCPHeader) )
+          return;
+        tcpheader = (struct TCPHeader *) ( packet +
+                                           sizeof (struct ICMPHeader) +
+                                           sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
+        break;
+#endif
+      }
+      sequence = ntohs(tcpheader->srcport);
+    }
+    break;
   }
+
+  if (sequence)
+    net_process_ping (sequence, mpls, (void *) fromaddress, now);
 }
 
 
@@ -539,6 +877,15 @@ ip_t *net_addrs(int at, int i)
   return (ip_t *)&(host[at].addrs[i]);
 }
 
+void *net_mpls(int at)
+{
+  return (struct mplslen *)&(host[at].mplss);
+}
+
+void *net_mplss(int at, int i)
+{
+  return (struct mplslen *)&(host[at].mplss[i]);
+}
 
 int net_loss(int at) 
 {
@@ -588,14 +935,13 @@ int net_gmean(int at)
 int net_stdev(int at) 
 {
   if( host[at].returned > 1 ) {
-    return ( sqrt( host[at].var/(host[at].returned -1.0) ) );
+    return ( 1000.0 * sqrt( host[at].var/(host[at].returned -1.0) ) );
   } else {
     return( 0 );
   }
 }
 
 
-/* jitter stuff */
 int net_jitter(int at) 
 { 
   return (host[at].jitter); 
@@ -618,7 +964,6 @@ int net_jinta(int at)
 { 
   return (host[at].jinta); 
 }
-/* end jitter */
 
 
 int net_max(void) 
@@ -627,8 +972,7 @@ int net_max(void)
   int max;
 
   max = 0;
-  // replacedByMin
-  // for(at = 0; at < MaxHost-2; at++) {
+  /* for(at = 0; at < MaxHost-2; at++) { */
   for(at = 0; at < maxTTL-1; at++) {
     if ( addrcmp( (void *) &(host[at].addr),
                   (void *) remoteaddress, af ) == 0 ) {
@@ -643,14 +987,12 @@ int net_max(void)
 }
 
 
-/* add by Min (wonder its named net_min;-)) because of ttl stuff */
 int net_min (void) 
 {
   return ( fstTTL - 1 );
 }
 
 
-/* Added by Brian Casey December 1997 bcasey@imagiware.com*/
 int net_returned(int at) 
 { 
   return host[at].returned;
@@ -707,7 +1049,7 @@ int net_send_batch(void)
            As our detination range (in the example 10000) is much 
            smaller (reasonable packet sizes), and our rand() range much 
            larger, this effect is insignificant. Oh! That other formula
-           didn't work. -- REW */
+           didn't work. */
       packetsize = MINPACKET + rand () % (-cpacketsize - MINPACKET);
     } else {
       packetsize = cpacketsize;
@@ -729,19 +1071,19 @@ int net_send_batch(void)
 	but I don't remember why. It makes mtr stop skipping sections of unknown
 	hosts. Removed in 0.65. 
 	If the line proves neccesary, it should at least NOT trigger that line 
-	when host[i].addr == 0 -- REW */
+	when host[i].addr == 0 */
     if ( ( addrcmp( (void *) &(host[i].addr),
                     (void *) remoteaddress, af ) == 0 )
 	/* || (host[i].addr == host[batch_at].addr)  */)
       n_unknown = MaxHost; /* Make sure we drop into "we should restart" */
   }
 
-  if (	// success in reaching target
+  if (	/* success in reaching target */
      ( addrcmp( (void *) &(host[batch_at].addr),
                 (void *) remoteaddress, af ) == 0 ) ||
-      // fail in consecuitive MAX_UNKNOWN_HOSTS (firewall?)
+      /* fail in consecuitive MAX_UNKNOWN_HOSTS (firewall?) */
       (n_unknown > MAX_UNKNOWN_HOSTS) ||
-      // or reach limit 
+      /* or reach limit  */
       (batch_at >= maxTTL-1)) {
     numhosts = batch_at+1;
     batch_at = fstTTL - 1;
@@ -753,19 +1095,38 @@ int net_send_batch(void)
 }
 
 
+static void set_fd_flags(int fd)
+{
+#if defined(HAVE_FCNTL) && defined(FD_CLOEXEC)
+  int oldflags;
+
+  if (fd < 0) return; 
+
+  oldflags = fcntl(fd, F_GETFD);
+  if (oldflags == -1) {
+    perror("Couldn't get fd's flags");
+    return;
+  }
+  if (fcntl(fd, F_SETFD, oldflags | FD_CLOEXEC))
+    perror("Couldn't set fd's flags");
+#endif
+}
+
 int net_preopen(void) 
 {
   int trueopt = 1;
 
 #if !defined(IP_HDRINCL) && defined(IP_TOS) && defined(IP_TTL)
-  sendsock4 = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
+  sendsock4_icmp = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
+  sendsock4_udp = socket(AF_INET, SOCK_RAW, IPPROTO_UDP);
 #else
   sendsock4 = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
 #endif
   if (sendsock4 < 0) 
     return -1;
 #ifdef ENABLE_IPV6
-  sendsock6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+  sendsock6_icmp = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+  sendsock6_udp = socket(AF_INET6, SOCK_RAW, IPPROTO_UDP);
 #endif
 
 #ifdef IP_HDRINCL
@@ -780,14 +1141,48 @@ int net_preopen(void)
   recvsock4 = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
   if (recvsock4 < 0)
     return -1;
+  set_fd_flags(recvsock4);
 #ifdef ENABLE_IPV6
   recvsock6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+  if (recvsock6 >= 0)
+     set_fd_flags(recvsock6);
 #endif
 
   return 0;
 }
 
- 
+
+int net_selectsocket(void)
+{
+#if !defined(IP_HDRINCL) && defined(IP_TOS) && defined(IP_TTL)
+  switch ( mtrtype ) {
+  case IPPROTO_ICMP:
+    sendsock4 = sendsock4_icmp;
+    break;
+  case IPPROTO_UDP:
+    sendsock4 = sendsock4_udp;
+    break;
+  }
+#endif
+  if (sendsock4 < 0)
+    return -1;
+#ifdef ENABLE_IPV6
+  switch ( mtrtype ) {
+  case IPPROTO_ICMP:
+    sendsock6 = sendsock6_icmp;
+    break;
+  case IPPROTO_UDP:
+    sendsock6 = sendsock6_udp;
+    break;
+  }
+  if ((sendsock6 < 0) && (sendsock4 < 0))
+    return -1;
+#endif
+
+ return 0;
+}
+
+
 int net_open(struct hostent * host) 
 {
 #ifdef ENABLE_IPV6
@@ -901,6 +1296,10 @@ void net_reset(void)
   
   for (at = 0; at < MaxSequence; at++) {
     sequence[at].transit = 0;
+    if (sequence[at].socket > 0) {
+      close(sequence[at].socket);
+      sequence[at].socket = 0;
+    }
   }
 
   gettimeofday(&reset, NULL);
@@ -946,9 +1345,15 @@ int net_set_interfaceaddress (char *InterfaceAddress)
 
 void net_close(void)
 {
-  if (sendsock4 >= 0) close(sendsock4);
+  if (sendsock4 >= 0) {
+    close(sendsock4_icmp);
+    close(sendsock4_udp);
+  }
   if (recvsock4 >= 0) close(recvsock4);
-  if (sendsock6 >= 0) close(sendsock6);
+  if (sendsock6 >= 0) {
+    close(sendsock6_icmp);
+    close(sendsock6_udp);
+  }
   if (recvsock6 >= 0) close(recvsock6);
 }
 
@@ -1004,7 +1409,7 @@ void sockaddrtop( struct sockaddr * saddr, char * strptr, size_t len ) {
   switch ( saddr->sa_family ) {
   case AF_INET:
     sa4 = (struct sockaddr_in *) saddr;
-    strncpy( strptr, inet_ntoa( (struct in_addr) sa4->sin_addr ), len - 1 );
+    strncpy( strptr, inet_ntoa( sa4->sin_addr ), len - 1 );
     strptr[ len - 1 ] = '\0';
     return;
 #ifdef ENABLE_IPV6
@@ -1052,3 +1457,109 @@ void addrcpy( char * a, char * b, int af ) {
 #endif
   }
 }
+
+/* Decode MPLS */
+void decodempls(int num, char *packet, struct mplslen *mpls, int offset) {
+
+  int i;
+  unsigned int ext_ver, ext_res, ext_chk, obj_hdr_len;
+  u_char obj_hdr_class, obj_hdr_type;
+
+  /* loosely derived from the traceroute-nanog.c
+   * decoding by Jorge Boncompte */
+  ext_ver = packet[offset]>>4;
+  ext_res = (packet[offset]&15)+ packet[offset+1];
+  ext_chk = ((unsigned int)packet[offset+2]<<8)+packet[offset+3];
+
+  /* Check for ICMP extension header */
+  if (ext_ver == 2 && ext_res == 0 && ext_chk != 0 && num >= (offset+6)) {
+    obj_hdr_len = ((int)packet[offset+4]<<8)+packet[offset+5];
+    obj_hdr_class = packet[offset+6];
+    obj_hdr_type = packet[offset+7];
+
+    /* make sure we have an MPLS extension */
+    if (obj_hdr_len >= 8 && obj_hdr_class == 1 && obj_hdr_type == 1) {
+      /* how many labels do we have?  will be at least 1 */
+      mpls->labels = (obj_hdr_len-4)/4;
+
+      /* save all label objects */
+      for(i=0; (i<mpls->labels) && (i < MAXLABELS) && (num >= (offset+8)+(i*4)); i++) {
+
+        /* piece together the 20 byte label value */
+        mpls->label[i] = ((unsigned long) (packet[(offset+8)+(i*4)] << 12 & 0xff000) +
+            (unsigned int) (packet[(offset+9)+(i*4)] << 4 & 0xff0) +
+            (packet[(offset+10)+(i*4)] >> 4 & 0xf));
+        mpls->exp[i] = (packet[(offset+10)+(i*4)] >> 1) & 0x7;
+        mpls->s[i] = (packet[(offset+10)+(i*4)] & 0x1); /* should be 1 if only one label */
+        mpls->ttl[i] = packet[(offset+11)+(i*4)];
+      }
+    }
+  }
+}
+
+/* Add open sockets to select() */
+void net_add_fds(fd_set *writefd, int *maxfd)
+{
+  int at, fd;
+  for (at = 0; at < MaxSequence; at++) {
+    fd = sequence[at].socket;
+    if (fd > 0) {
+      FD_SET(fd, writefd);
+      if (fd >= *maxfd)
+        *maxfd = fd + 1;
+    }
+  }
+}
+
+/* check if we got connection or error on any fds */
+void net_process_fds(fd_set *writefd)
+{
+  int at, fd, r;
+  struct timeval now;
+  uint64_t unow, utime;
+
+  /* Can't do MPLS decoding */
+  struct mplslen mpls;
+  mpls.labels = 0;
+
+  gettimeofday(&now, NULL);
+  unow = now.tv_sec * 1000000L + now.tv_usec;
+
+  for (at = 0; at < MaxSequence; at++) {
+    fd = sequence[at].socket;
+    if (fd > 0 && FD_ISSET(fd, writefd)) {
+      r = write(fd, "G", 1);
+      /* if write was successful, or connection refused we have
+       * (probably) reached the remote address. Anything else happens to the
+       * connection, we write it off to avoid leaking sockets */
+      if (r == 1 || errno == ECONNREFUSED)
+        net_process_ping(at, mpls, remoteaddress, now);
+      else if (errno != EAGAIN) {
+        close(fd);
+        sequence[at].socket = 0;
+      }
+    }
+    if (fd > 0) {
+      utime = sequence[at].time.tv_sec * 1000000L + sequence[at].time.tv_usec;
+      if (unow - utime > timeout) {
+        close(fd);
+        sequence[at].socket = 0;
+      }
+    }
+  }
+}
+
+/* for GTK frontend */
+void net_harvest_fds(void)
+{
+  fd_set writefd;
+  int maxfd = 0;
+  struct timeval tv;
+
+  FD_ZERO(&writefd);
+  tv.tv_sec = 0;
+  tv.tv_usec = 0;
+  net_add_fds(&writefd, &maxfd);
+  select(maxfd, NULL, &writefd, NULL, &tv);
+  net_process_fds(&writefd);
+}
